@@ -16,7 +16,7 @@
     _PREFS_FILE */
 
 /* defined by XPCShellImpl.cpp */
-/* globals load, sendCommand */
+/* globals load, sendCommand, changeTestShellDir */
 
 /* must be defined by tests using do_await_remote_message/do_send_remote_message */
 /* globals Cc, Ci */
@@ -34,43 +34,38 @@ var _tests_pending = 0;
 var _cleanupFunctions = [];
 var _pendingTimers = [];
 var _profileInitialized = false;
-var _fastShutdownDisabled = false;
 
 // Assigned in do_load_child_test_harness.
 var _XPCSHELL_PROCESS;
 
 // Register the testing-common resource protocol early, to have access to its
 // modules.
-let { Services: _Services } = ChromeUtils.import(
-  "resource://gre/modules/Services.jsm"
-);
+let _Services = Services;
 _register_modules_protocol_handler();
 
-let { AppConstants: _AppConstants } = ChromeUtils.import(
-  "resource://gre/modules/AppConstants.jsm"
+let { AppConstants: _AppConstants } = ChromeUtils.importESModule(
+  "resource://gre/modules/AppConstants.sys.mjs"
 );
 
-let { PromiseTestUtils: _PromiseTestUtils } = ChromeUtils.import(
-  "resource://testing-common/PromiseTestUtils.jsm"
+let { PromiseTestUtils: _PromiseTestUtils } = ChromeUtils.importESModule(
+  "resource://testing-common/PromiseTestUtils.sys.mjs"
 );
 
-let { NetUtil: _NetUtil } = ChromeUtils.import(
-  "resource://gre/modules/NetUtil.jsm"
+let { NetUtil: _NetUtil } = ChromeUtils.importESModule(
+  "resource://gre/modules/NetUtil.sys.mjs"
 );
 
-let { XPCOMUtils: _XPCOMUtils } = ChromeUtils.import(
-  "resource://gre/modules/XPCOMUtils.jsm"
+let { XPCOMUtils: _XPCOMUtils } = ChromeUtils.importESModule(
+  "resource://gre/modules/XPCOMUtils.sys.mjs"
 );
 
-let { OS: _OS } = ChromeUtils.import("resource://gre/modules/osfile.jsm");
-
-// Support a common assertion library, Assert.jsm.
-var { Assert: AssertCls } = ChromeUtils.import(
-  "resource://testing-common/Assert.jsm"
+// Support a common assertion library, Assert.sys.mjs.
+var { Assert: AssertCls } = ChromeUtils.importESModule(
+  "resource://testing-common/Assert.sys.mjs"
 );
 
 // Pass a custom report function for xpcshell-test style reporting.
-var Assert = new AssertCls(function(err, message, stack) {
+var Assert = new AssertCls(function (err, message, stack) {
   if (err) {
     do_report_result(false, err.message, err.stack);
   } else {
@@ -78,24 +73,24 @@ var Assert = new AssertCls(function(err, message, stack) {
   }
 }, true);
 
-// Bug 1506134 for followup.  Some xpcshell tests use ContentTask.jsm, which
+// Bug 1506134 for followup.  Some xpcshell tests use ContentTask.sys.mjs, which
 // expects browser-test.js to have set a testScope that includes record.
 function record(condition, name, diag, stack) {
   do_report_result(condition, name, stack);
 }
 
-var _add_params = function(params) {
+var _add_params = function (params) {
   if (typeof _XPCSHELL_PROCESS != "undefined") {
     params.xpcshell_process = _XPCSHELL_PROCESS;
   }
 };
 
-var _dumpLog = function(raw_msg) {
+var _dumpLog = function (raw_msg) {
   dump("\n" + JSON.stringify(raw_msg) + "\n");
 };
 
-var { StructuredLogger: _LoggerClass } = ChromeUtils.import(
-  "resource://testing-common/StructuredLog.jsm"
+var { StructuredLogger: _LoggerClass } = ChromeUtils.importESModule(
+  "resource://testing-common/StructuredLog.sys.mjs"
 );
 var _testLogger = new _LoggerClass("xpcshell/head.js", _dumpLog, [_add_params]);
 
@@ -129,11 +124,21 @@ if (runningInParent && "mozIAsyncHistory" in Ci) {
 // crashreporter component.
 try {
   if (runningInParent && "@mozilla.org/toolkit/crash-reporter;1" in Cc) {
+    // Intentially access the crash reporter service directly for this.
+    // eslint-disable-next-line mozilla/use-services
     let crashReporter = Cc["@mozilla.org/toolkit/crash-reporter;1"].getService(
       Ci.nsICrashReporter
     );
     crashReporter.UpdateCrashEventsDir();
     crashReporter.minidumpPath = do_get_minidumpdir();
+
+    // Tell the test harness that the crash reporter is set up, and any crash
+    // after this point is supposed to be caught by the crash reporter.
+    //
+    // Due to the limitation on the remote xpcshell test, the process return
+    // code does not represent the process crash. Any crash before this point
+    // can be caught by the absence of this log.
+    _testLogger.logData("crash_reporter_init");
   }
 } catch (e) {}
 
@@ -390,20 +395,17 @@ function _setupDevToolsServer(breakpointFiles, callback) {
   _Services.prefs.setBoolPref("devtools.debugger.remote-enabled", true);
 
   // for debugging-the-debugging, let an env var cause log spew.
-  let env = Cc["@mozilla.org/process/environment;1"].getService(
-    Ci.nsIEnvironment
-  );
-  if (env.get("DEVTOOLS_DEBUGGER_LOG")) {
+  if (_Services.env.get("DEVTOOLS_DEBUGGER_LOG")) {
     _Services.prefs.setBoolPref("devtools.debugger.log", true);
   }
-  if (env.get("DEVTOOLS_DEBUGGER_LOG_VERBOSE")) {
+  if (_Services.env.get("DEVTOOLS_DEBUGGER_LOG_VERBOSE")) {
     _Services.prefs.setBoolPref("devtools.debugger.log.verbose", true);
   }
 
   let require;
   try {
-    ({ require } = ChromeUtils.import(
-      "resource://devtools/shared/loader/Loader.jsm"
+    ({ require } = ChromeUtils.importESModule(
+      "resource://devtools/shared/loader/Loader.sys.mjs"
     ));
   } catch (e) {
     throw new Error(
@@ -411,7 +413,7 @@ function _setupDevToolsServer(breakpointFiles, callback) {
         "xpcshell environment.\n" +
         "This can usually be resolved by adding:\n" +
         "  firefox-appdir = browser\n" +
-        "to the xpcshell.ini manifest.\n" +
+        "to the xpcshell.toml manifest.\n" +
         "It is possible for this to alter test behevior by " +
         "triggering additional browser code to run, so check " +
         "test behavior after making this change.\n" +
@@ -432,7 +434,7 @@ function _setupDevToolsServer(breakpointFiles, callback) {
     // Or when devtools are destroyed and we should stop observing.
     "xpcshell-test-devtools-shutdown",
   ];
-  let observe = function(subject, topic, data) {
+  let observe = function (subject, topic) {
     if (topic === "devtools-thread-ready") {
       const threadActor = subject.wrappedJSObject;
       threadActor.setBreakpointOnLoad(breakpointFiles);
@@ -486,11 +488,15 @@ function _initDebugging(port) {
 
   // spin an event loop until the debugger connects.
   const tm = Cc["@mozilla.org/thread-manager;1"].getService();
+  let lastUpdate = Date.now();
   tm.spinEventLoopUntil("Test(xpcshell/head.js:_initDebugging)", () => {
     if (initialized) {
       return true;
     }
-    info("Still waiting for debugger to connect...");
+    if (Date.now() - lastUpdate > 5000) {
+      info("Still waiting for debugger to connect...");
+      lastUpdate = Date.now();
+    }
     return false;
   });
   // NOTE: if you want to debug the harness itself, you can now add a 'debugger'
@@ -501,19 +507,12 @@ function _initDebugging(port) {
 
 function _execute_test() {
   if (typeof _TEST_CWD != "undefined") {
-    let cwd_complete = false;
-    _OS.File.setCurrentDirectory(_TEST_CWD)
-      .then(_ => (cwd_complete = true))
-      .catch(e => {
-        _testLogger.error(_exception_message(e));
-        cwd_complete = true;
-      });
-    _Services.tm.spinEventLoopUntil(
-      "Test(xpcshell/head.js:setCurrentDirectory)",
-      () => cwd_complete
-    );
+    try {
+      changeTestShellDir(_TEST_CWD);
+    } catch (e) {
+      _testLogger.error(_exception_message(e));
+    }
   }
-
   if (runningInParent && _AppConstants.platform == "android") {
     try {
       // GeckoView initialization needs the profile
@@ -555,8 +554,8 @@ function _execute_test() {
 
   let coverageCollector = null;
   if (typeof _JSCOV_DIR === "string") {
-    let _CoverageCollector = ChromeUtils.import(
-      "resource://testing-common/CoverageUtils.jsm"
+    let _CoverageCollector = ChromeUtils.importESModule(
+      "resource://testing-common/CoverageUtils.sys.mjs"
     ).CoverageCollector;
     coverageCollector = new _CoverageCollector(_JSCOV_DIR);
   }
@@ -568,14 +567,14 @@ function _execute_test() {
   // _TEST_FILE is dynamically defined by <runxpcshelltests.py>.
   _load_files(_TEST_FILE);
 
-  // Tack Assert.jsm methods to the current scope.
+  // Tack Assert.sys.mjs methods to the current scope.
   this.Assert = Assert;
   for (let func in Assert) {
     this[func] = Assert[func].bind(Assert);
   }
 
-  const { PerTestCoverageUtils } = ChromeUtils.import(
-    "resource://testing-common/PerTestCoverageUtils.jsm"
+  const { PerTestCoverageUtils } = ChromeUtils.importESModule(
+    "resource://testing-common/PerTestCoverageUtils.sys.mjs"
   );
 
   if (runningInParent) {
@@ -635,7 +634,7 @@ function _execute_test() {
   }
 
   // Execute all of our cleanup functions.
-  let reportCleanupError = function(ex) {
+  let reportCleanupError = function (ex) {
     let stack, filename;
     if (ex && typeof ex == "object" && "stack" in ex) {
       stack = ex.stack;
@@ -653,7 +652,7 @@ function _execute_test() {
     });
   };
 
-  let complete = _cleanupFunctions.length == 0;
+  let complete = !_cleanupFunctions.length;
   let cleanupStartTime = complete ? 0 : Cu.now();
   (async () => {
     for (let func of _cleanupFunctions.reverse()) {
@@ -692,6 +691,13 @@ function _execute_test() {
   // Restore idle service to avoid leaks.
   _fakeIdleService.deactivate();
 
+  if (
+    globalThis.hasOwnProperty("storage") &&
+    StorageManager.isInstance(globalThis.storage)
+  ) {
+    globalThis.storage.shutdown();
+  }
+
   if (_profileInitialized) {
     // Since we have a profile, we will notify profile shutdown topics at
     // the end of the current test, to ensure correct cleanup on shutdown.
@@ -729,16 +735,6 @@ function _execute_test() {
     !_AppConstants.ASAN &&
     !_AppConstants.TSAN
   ) {
-    if (_fastShutdownDisabled) {
-      _testLogger.info("fast shutdown disabled by the test.");
-      return;
-    }
-
-    // Setting this pref is required for Cu.isInAutomation to return true.
-    _Services.prefs.setBoolPref(
-      "security.turn_off_all_security_so_that_viruses_can_take_over_this_computer",
-      true
-    );
     Cu.exitIfInAutomation();
   }
 }
@@ -749,7 +745,7 @@ function _execute_test() {
  * @param aFiles Array of files to load.
  */
 function _load_files(aFiles) {
-  function load_file(element, index, array) {
+  function load_file(element) {
     try {
       let startTime = Cu.now();
       load(element);
@@ -858,6 +854,11 @@ function do_throw(error, stack) {
     source_file: filename,
     stack: _format_stack(stack),
   });
+  ChromeUtils.addProfilerMarker(
+    "ERROR",
+    { category: "Test", captureStack: true },
+    _exception_message(error)
+  );
   _abort_failed_test();
 }
 
@@ -885,6 +886,9 @@ function _format_stack(stack) {
 // Make a nice display string from an object that behaves
 // like Error
 function _exception_message(ex) {
+  if (ex === undefined) {
+    return "`undefined` exception, maybe from an empty reject()?";
+  }
   let message = "";
   if (ex.name) {
     message = ex.name + ": ";
@@ -912,7 +916,7 @@ function do_report_unexpected_exception(ex, text) {
   _passed = false;
   _testLogger.error(text + "Unexpected exception " + _exception_message(ex), {
     source_file: filename,
-    stack: _format_stack(ex.stack),
+    stack: _format_stack(ex?.stack),
   });
   _do_quit();
   throw Components.Exception("", Cr.NS_ERROR_ABORT);
@@ -922,7 +926,7 @@ function do_note_exception(ex, text) {
   let filename = Components.stack.caller.filename;
   _testLogger.info(text + "Swallowed exception " + _exception_message(ex), {
     source_file: filename,
-    stack: _format_stack(ex.stack),
+    stack: _format_stack(ex?.stack),
   });
 }
 
@@ -951,12 +955,19 @@ function do_report_result(passed, text, stack, todo) {
         message,
         _format_stack(stack)
       );
+      ChromeUtils.addProfilerMarker(
+        "UNEXPECTED-PASS",
+        { category: "Test" },
+        message
+      );
       _abort_failed_test();
     } else {
       _testLogger.testStatus(_TEST_NAME, name, "PASS", "PASS", message);
+      ChromeUtils.addProfilerMarker("PASS", { category: "Test" }, message);
     }
   } else if (todo) {
     _testLogger.testStatus(_TEST_NAME, name, "FAIL", "FAIL", message);
+    ChromeUtils.addProfilerMarker("TODO", { category: "Test" }, message);
   } else {
     _testLogger.testStatus(
       _TEST_NAME,
@@ -966,6 +977,7 @@ function do_report_result(passed, text, stack, todo) {
       message,
       _format_stack(stack)
     );
+    ChromeUtils.addProfilerMarker("FAIL", { category: "Test" }, message);
     _abort_failed_test();
   }
 }
@@ -1226,14 +1238,6 @@ function registerCleanupFunction(aFunction) {
   _cleanupFunctions.push(aFunction);
 }
 
-/**
- * Ensure the test finishes with a normal shutdown even when it could have
- * otherwise used the fast Cu.exitIfInAutomation shutdown.
- */
-function do_disable_fast_shutdown() {
-  _fastShutdownDisabled = true;
-}
-
 /**
  * Returns the directory for a temp dir, which is created by the
  * test harness. Every test gets its own temp dir.
@@ -1241,11 +1245,8 @@ function do_disable_fast_shutdown() {
  * @return nsIFile of the temporary directory
  */
 function do_get_tempdir() {
-  let env = Cc["@mozilla.org/process/environment;1"].getService(
-    Ci.nsIEnvironment
-  );
   // the python harness sets this in the environment for us
-  let path = env.get("XPCSHELL_TEST_TEMP_DIR");
+  let path = _Services.env.get("XPCSHELL_TEST_TEMP_DIR");
   let file = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);
   file.initWithPath(path);
   return file;
@@ -1257,11 +1258,8 @@ function do_get_tempdir() {
  * @return nsIFile of the minidump directory
  */
 function do_get_minidumpdir() {
-  let env = Cc["@mozilla.org/process/environment;1"].getService(
-    Ci.nsIEnvironment
-  );
   // the python harness may set this in the environment for us
-  let path = env.get("XPCSHELL_MINIDUMP_DIR");
+  let path = _Services.env.get("XPCSHELL_MINIDUMP_DIR");
   if (path) {
     let file = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);
     file.initWithPath(path);
@@ -1283,11 +1281,8 @@ function do_get_profile(notifyProfileAfterChange = false) {
     return null;
   }
 
-  let env = Cc["@mozilla.org/process/environment;1"].getService(
-    Ci.nsIEnvironment
-  );
   // the python harness sets this in the environment for us
-  let profd = env.get("XPCSHELL_TEST_PROFILE_DIR");
+  let profd = Services.env.get("XPCSHELL_TEST_PROFILE_DIR");
   let file = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);
   file.initWithPath(profd);
 
@@ -1323,6 +1318,8 @@ function do_get_profile(notifyProfileAfterChange = false) {
 
   // We need to update the crash events directory when the profile changes.
   if (runningInParent && "@mozilla.org/toolkit/crash-reporter;1" in Cc) {
+    // Intentially access the crash reporter service directly for this.
+    // eslint-disable-next-line mozilla/use-services
     let crashReporter = Cc["@mozilla.org/toolkit/crash-reporter;1"].getService(
       Ci.nsICrashReporter
     );
@@ -1347,7 +1344,6 @@ function do_get_profile(notifyProfileAfterChange = false) {
 
   // The methods of 'provider' will retain this scope so null out everything
   // to avoid spurious leak reports.
-  env = null;
   profd = null;
   provider = null;
   return file.clone();
@@ -1571,7 +1567,7 @@ function add_test(
 }
 
 /**
- * Add a test function which is a Task function.
+ * Add a test function which is an asynchronous function.
  *
  * @param funcOrProperties
  *        An async function to be run or an object represents test properties.
@@ -1582,9 +1578,6 @@ function add_test(
  * @param func
  *        An async function to be run only if the funcOrProperies is not a function.
  *
- * Task functions are functions fed into Task.jsm's Task.spawn(). They are async
- * functions that emit promises.
- *
  * If an exception is thrown, a do_check_* comparison fails, or if a rejected
  * promise is yielded, the test function aborts immediately and the test is
  * reported as a failure.
@@ -1786,10 +1779,30 @@ function run_next_test() {
             );
             _setTaskPrefs(initialPrefsValues);
             try {
+              // Note `ex` at this point could be undefined, for example as
+              // result of a bare call to reject().
               do_report_unexpected_exception(ex);
             } catch (error) {
-              // The above throws NS_ERROR_ABORT and we don't want this to show up
-              // as an unhandled rejection later.
+              // The above throws NS_ERROR_ABORT and we don't want this to show
+              // up as an unhandled rejection later. If any other exception
+              // happened, something went wrong, so we abort.
+              if (error.result != Cr.NS_ERROR_ABORT) {
+                let extra = {};
+                if (error.fileName) {
+                  extra.source_file = error.fileName;
+                  if (error.lineNumber) {
+                    extra.line_number = error.lineNumber;
+                  }
+                } else {
+                  extra.source_file = "xpcshell/head.js";
+                }
+                if (error.stack) {
+                  extra.stack = _format_stack(error.stack);
+                }
+                _testLogger.error(_exception_message(error), extra);
+                _do_quit();
+                throw Components.Exception("", Cr.NS_ERROR_ABORT);
+              }
             }
           }
         );
@@ -1813,7 +1826,7 @@ function run_next_test() {
   }
 
   function frontLoadSetups() {
-    _gTests.sort(([propsA, funcA], [propsB, funcB]) => {
+    _gTests.sort(([propsA], [propsB]) => {
       if (propsB.isSetup === propsA.isSetup) {
         return 0;
       }
@@ -1860,9 +1873,10 @@ try {
   // We only want to run this for local developer builds (which should have a "default" update channel).
   if (runningInParent && _AppConstants.MOZ_UPDATE_CHANNEL == "default") {
     let startTime = Cu.now();
-    let { TelemetryController: _TelemetryController } = ChromeUtils.import(
-      "resource://gre/modules/TelemetryController.jsm"
-    );
+    let { TelemetryController: _TelemetryController } =
+      ChromeUtils.importESModule(
+        "resource://gre/modules/TelemetryController.sys.mjs"
+      );
 
     let complete = false;
     _TelemetryController.testRegisterJsProbes().finally(() => {
